@@ -25,7 +25,7 @@ header files or Java imports. It can deal with
 It counts
 
 * the nloc (net lines of code, excluding comments and blanks),
-* CCN (cyclomatic complexity number),
+* CCN (cyclomatic complexity number) or Modified CCN,
 * token count of functions.
 * parameter count of functions.
 
@@ -44,7 +44,7 @@ It requires python2.6 or above (early versions are not verified).
 
 BUG_REPORTING = "please report bug to terry.yinzhe@gmail.com or https://github.com/terryyin/lizard.\n"
 
-VERSION = "1.7.5"
+VERSION = "1.7.6"
 
 
 import itertools, traceback
@@ -54,14 +54,83 @@ import re
 DEFAULT_CCN_THRESHOLD = 15
 
 
-'''
-The analyzing process
-=====================
-Source code => Tokens => SourceCodeInforamtionBuilder
+def createCommandLineParser():
+    from optparse import OptionParser
+    parser = OptionParser(version=VERSION)
+    parser.add_option("-v", "--verbose",
+            help="Output in verbose mode (long function name)",
+            action="store_true",
+            dest="verbose",
+            default=False)
+    parser.add_option("-C", "--CCN",
+            help =  "Threshold for cyclomatic complexity number warning. "+
+                    "The default value is %d. Functions with CCN bigger than this number will generate warning" % DEFAULT_CCN_THRESHOLD,
+            action="store",
+            type="int",
+            dest="CCN",
+            default=DEFAULT_CCN_THRESHOLD)
+    parser.add_option("-a", "--arguments",
+            help="Limit for number of parameters",
+            action="store",
+            type="int",
+            dest="arguments",
+            default=100)
+    parser.add_option("-w", "--warnings_only",
+            help="Show warnings only, using clang/gcc's warning format for printing warnings. http://clang.llvm.org/docs/UsersManual.html#cmdoption-fdiagnostics-format",
+            action="store_true",
+            dest="warnings_only",
+            default=False)
+    parser.add_option("-i", "--ignore_warnings",
+            help="If the number of warnings is equal or less than the number, the tool will exit normally, otherwize it will generate error. Useful in makefile when improving legacy code.",
+            action="store",
+            type="int",
+            dest="number",
+            default=0)
+    parser.add_option("-x", "--exclude",
+            help="Exclude files that match this pattern. * matches everything, ? matches any single characoter, \"./folder/*\" exclude everything in the folder, recursively. Multiple patterns can be specified. Don't forget to add \"\" around the pattern.",
+            action="append",
+            dest="exclude",
+            default=[])
+    parser.add_option("-X", "--xml",
+            help="Generate XML in cppncss style instead of the normal tabular output. Useful to generate report in Jenkins server",
+            action="store_true",
+            dest="xml",
+            default=None)
+    parser.add_option("-P", "--no_preprocessor_count",
+            help="By default, a #if will also increase the complexity. Adding this option to ignore them",
+            action="store_true",
+            dest="no_preprocessor_count",
+            default=False)
+    parser.add_option("-t", "--working_threads",
+            help="number of working threads. The default value is 1.",
+            action="store",
+            type="int",
+            dest="working_threads",
+            default=1)
+    parser.add_option("-d", "--find_duplicates",
+            help="find and skip analysis for identical files. Will be made default in the next release",
+            action="store_true",
+            dest="duplicates",
+            default=False)
+    parser.add_option("-e", "--display_fn_end_line",
+            help="display function end line number in addition to start line number. Will be made default in the next release",
+            action="store_true",
+            dest="display_fn_end_line",
+            default=False)
+    parser.add_option("-m", "--modified",
+            help="Calculate modified cyclomatic complexity number",
+            action="store_true",
+            dest="switchCasesAsOneCondition",
+            default=False)
+    parser.add_option("-E", "--extension",
+            help="under construction...", #"Use extension. Can be WordCount.",
+            action="append",
+            dest="extensions",
+            default=[])
 
-SourceCodeInforamtionBuilder.get_statistics => The statistics of one source file.
-
-'''
+    parser.usage = "lizard [options] [PATH or FILE] [PATH] ... "
+    parser.description = __doc__
+    return parser
 
 
 class FunctionInfo(object):
@@ -130,16 +199,36 @@ class SourceCodeInforamtionBuilder(object):
         END_OF_FUNCTION
     """
 
-    def __init__(self):
-        self.fileinfo = FileInformation("", 0, [])
-        self.START_NEW_FUNCTION('', 0)
+    def __init__(self, filename):
+        self.fileinfo = FileInformation(filename, 0, [])
+        self.current_line = 0
+        self.START_NEW_FUNCTION('')
 
-    def build(self):
+    def parse_functions(self, tokens, parser):
+        parser.sourceCodeInfoBuilder = self
+        self.current_line = 1 
+        for token in tokens:
+            self.current_line += 1 if token == '\n' else (len(token.splitlines()) - 1)
+            try:
+                if token.isspace():
+                    self.NEW_LINE()
+                elif token.startswith("/*") or token.startswith("//"):
+                    if len(token.splitlines()) > 1:
+                        self.NEW_LINE()
+                    if token[2:].strip().startswith("#lizard forgive"):
+                        self.forgive = True
+                else:
+                    self.TOKEN()
+                    self.fileinfo.token_count+=1
+                    parser.process_token(token)
+            except:
+                raise ParsingError(self.current_line)
         return self.fileinfo
 
-    def START_NEW_FUNCTION(self, name, start_line):
+    def START_NEW_FUNCTION(self, name):
+        self.forgive = False
         self.newline = True
-        self.current_function = FunctionInfo(name, start_line)
+        self.current_function = FunctionInfo(name, self.current_line)
 
     def CONDITION(self):
         self.current_function.cyclomatic_complexity += 1
@@ -163,10 +252,11 @@ class SourceCodeInforamtionBuilder(object):
     def PARAMETER(self, token):
         self.current_function.add_parameter(token)
 
-    def END_OF_FUNCTION(self, end_line):
-        self.current_function.end_line = end_line
-        self.fileinfo.function_list.append(self.current_function)
-        self.START_NEW_FUNCTION('', 0)
+    def END_OF_FUNCTION(self):
+        self.current_function.end_line = self.current_line
+        if not self.forgive:
+            self.fileinfo.function_list.append(self.current_function)
+        self.START_NEW_FUNCTION('')
 
 
 class ParsingError(Exception):
@@ -202,22 +292,6 @@ class LanguageReaderBase(object):
 
     def __init__(self):
         self._state = self._GLOBAL
-        self.current_line = 0
-        self.sourceCodeInfoBuilder = SourceCodeInforamtionBuilder()
-
-    def generate_universal_code(self, tokens):
-        for token, self.current_line in tokens:
-            try:
-                if token.isspace():
-                    self.sourceCodeInfoBuilder.NEW_LINE()
-                else:
-                    self.sourceCodeInfoBuilder.TOKEN()
-                    self.sourceCodeInfoBuilder.fileinfo.token_count+=1
-                    self.process_token(token)
-            except:
-                raise ParsingError(self.current_line)
-
-        return self.sourceCodeInfoBuilder
 
     def process_token(self, token):
         return self._state(token)
@@ -231,34 +305,42 @@ class CLikeReader(LanguageReaderBase):
     This is the reader for C, C++ and Java.
     '''
 
-    def __init__(self):
-        super(CLikeReader, self).__init__()
+    def __init__(self, includeHashIfConditions = True, switchCasesAsOneCondition = False, **kwargs):
+        '''includeHashIfConditions: Should #if and #elseif have effect on the CCN
+           switchCasesAsOneCondition: Use modified CCN (i.e. for one switch case increases CCN by 1 instead of
+                        by the amount of case blocks)
+        '''
+        super(CLikeReader, self).__init__(**kwargs)
         self.conditions = set(
-            ['if', 'for', 'while', '&&', '||', 'case', '?', '#if', '#elif', 'catch'])
+            ['if', 'for', 'while', '&&', '||', '?', 'catch'])
+
+        if includeHashIfConditions:
+            self.conditions.add('#if')
+            self.conditions.add('#ifdef')
+            self.conditions.add('#elif')
+
+        if switchCasesAsOneCondition:
+            self.conditions.add('switch')
+        else:
+            self.conditions.add('case')
+
         self.bracket_level = 0
         self.br_count = 0
         self.last_preprocessor = None
 
-    def extend_tokens(self, tokens):
-        last_token_is_include = False
-        include_brackets = False
-        brackets = ""
-        for token, line in tokens:
-            if last_token_is_include and token == "<":
-                include_brackets = True
-                brackets = "<"
-            elif include_brackets:
-                brackets += token
-                if token == ">":
-                    include_brackets = False
-                    token = brackets
-            last_token_is_include = token == "#include"
-            if not include_brackets:
-                yield token, line
+    macro_pattern = re.compile(r"#\s*(\w+)\s*(.*)", re.M | re.S)
 
-    def remove_hash_if_from_conditions(self):
-        self.conditions.remove("#if")
-        self.conditions.remove("#elif")
+    def extend_tokens(self, tokens):
+        # handle c preprocessors
+        for token in tokens:
+            m = self.macro_pattern.match(token)
+            if m:
+                yield "#" + m.group(1)
+                param = m.group(2).strip()
+                if param:
+                    yield param
+            else:
+                yield token
 
     def _is_condition(self, token):
         return token in self.conditions
@@ -276,7 +358,7 @@ class CLikeReader(LanguageReaderBase):
         elif token == '::':
             self._state = self._NAMESPACE
         else:
-            self.sourceCodeInfoBuilder.START_NEW_FUNCTION(token, self.current_line)
+            self.sourceCodeInfoBuilder.START_NEW_FUNCTION(token)
             if token == 'operator':
                 self._state = self._OPERATOR
 
@@ -318,7 +400,7 @@ class CLikeReader(LanguageReaderBase):
             self._state = self._IMP
 
     def _IMP(self, token):
-        if token in ("#else", "#if", "#endif"):
+        if token in ("#else", "#endif"):
             self.last_preprocessor = token
         # will ignore the braces in a #else branch            
         if self.last_preprocessor != '#else':
@@ -328,15 +410,15 @@ class CLikeReader(LanguageReaderBase):
                 self.br_count -= 1
                 if self.br_count == 0:
                     self._state = self._GLOBAL
-                    self.sourceCodeInfoBuilder.END_OF_FUNCTION(self.current_line)
+                    self.sourceCodeInfoBuilder.END_OF_FUNCTION()
                     return
         if self._is_condition(token):
             self.sourceCodeInfoBuilder.CONDITION()
 
 
 class ObjCReader(CLikeReader):
-    def __init__(self):
-        super(ObjCReader, self).__init__()
+    def __init__(self, **kwargs):
+        super(ObjCReader, self).__init__(**kwargs)
 
     def _DEC_TO_IMP(self, token):
         if token in ("+", "-"):
@@ -345,7 +427,7 @@ class ObjCReader(CLikeReader):
             super(ObjCReader, self)._DEC_TO_IMP(token)
             if self._state == self._GLOBAL:
                 self._state = self._OBJC_DEC_BEGIN
-                self.sourceCodeInfoBuilder.START_NEW_FUNCTION(token, self.current_line)
+                self.sourceCodeInfoBuilder.START_NEW_FUNCTION(token)
 
     def _OBJC_DEC_BEGIN(self, token):
         if token == ':':
@@ -384,6 +466,10 @@ def compile_file_extension_re(*exts):
 
 class LanguageChooser(object):
 
+    def __init__(self, **readerArgs):
+        object.__init__(self)
+        self.readerArgs = readerArgs
+
     lizard_language_infos = {
                      'c/c++': {
                           'name_pattern': compile_file_extension_re("c", "cpp", "cc", "mm", "cxx", "h", "hpp"),
@@ -404,17 +490,21 @@ class LanguageChooser(object):
 
     def get_reader_by_file_name_otherwise_default(self, filename):
         lan = self.get_language_by_filename(filename)
-        return self.lizard_language_infos[lan or "c/c++"]['reader']()
+        return self.lizard_language_infos[lan or "c/c++"]['reader'](**self.readerArgs)
 
+default_language_chooser = LanguageChooser(
+    includeHashIfConditions = True,
+    switchCasesAsOneCondition = False,
+)
 
 class FileAnalyzer:
     ''' A FileAnalyzer works as a function. It takes filename as parameter.
         Returns a list of function infos in this file.
     '''
 
-    def __init__(self, extensions = [], noCountPre=False):
-        self.no_preprocessor_count = noCountPre
+    def __init__(self, languageChooser = default_language_chooser, extensions = []):
         self.extensions = extensions
+        self.languageChooser = languageChooser
 
     def __call__(self, filename):
         try:
@@ -427,12 +517,8 @@ class FileAnalyzer:
 
     def analyze_source_code(self, filename, code):
         try:
-            reader = LanguageChooser().get_reader_by_file_name_otherwise_default(filename)
-            if self.no_preprocessor_count and hasattr(reader, 'remove_hash_if_from_conditions'):
-                reader.remove_hash_if_from_conditions()
-            result = self.analyze_source_code_with_parser(filename, code, reader)
-
-            return result
+            reader = self.languageChooser.get_reader_by_file_name_otherwise_default(filename)
+            return self.analyze_source_code_with_parser(filename, code, reader)
         except ParsingError as e:
             e.filename = filename
             e.code = code
@@ -443,8 +529,7 @@ class FileAnalyzer:
         tokens = parser.extend_tokens(tokens)
         for extension in self.extensions:
             tokens = extension.extend_tokens(tokens)
-        result = parser.generate_universal_code(tokens).build()
-        result.filename = filename
+        result = SourceCodeInforamtionBuilder(filename).parse_functions(tokens, parser)
         for extension in self.extensions:
             result.wordCount = extension.result
         return result
@@ -499,74 +584,23 @@ class FreeFormattingTokenizer(object):
     doesn't matter.
     '''
 
-    token_pattern = re.compile(r"(\w+|/\*|//|:=|::|>=|\*=|\*\*|\*|>"+
+    # DONOT put any sub groups in the regex. Good for performance
+    _until_end = r"(?:\\\n|[^\n])*"
+    token_pattern = re.compile(r"(\w+"+
+                           r"|/\*.*?\*/"+
+                           r"|\"(?:\\.|[^\"])*\""+
+                           r"|\'(?:\\.|[^\'])*\'"+
+                           r"|//"+ _until_end +
+                           r"|#" + _until_end +
+                           r"|:=|::|>=|\*=|\*\*|\*|>"+
                            r"|&=|&&|&"+
-                           r"|#\s*define|#\s*if|#\s*else|#\s*endif|#\s*\w+"+
-                           r"|[!%^&\*\-=+\|\\<>/\]\+]+|.)", re.M | re.S)
-
+                           r"|[!%^&\*\-=+\|\\<>/\]\+]+"+
+                           r"|\n"+
+                           r"|\s+"+
+                           r"|.)", re.M | re.S)
 
     def __call__(self, source_code):
-        for t, l in self._generate_tokens_without_empty_lines(source_code):
-            if not any(t.startswith(x) for x in ('#define', '/*', '//')) :
-                yield t, l
-
-    def _generate_tokens_without_empty_lines(self, source_code):
-        in_middle_of_empty_lines = False
-        for (token, line) in self._tokens_from_code_with_multiple_newlines(source_code):
-            if token != '\n' or not in_middle_of_empty_lines:
-                yield token, line
-            in_middle_of_empty_lines = (token == '\n')
-
-    def _tokens_from_code_with_multiple_newlines(self, source_code):
-        index = 0
-        line = 1
-        while index >= 0:
-            m = self.token_pattern.match(source_code, index)
-            if not m:
-                break
-            index, token = self._read_one_token(source_code, index, m.group(0))            
-            line += 1 if token == '\n' else (len(token.splitlines()) - 1)
-            if not token.isspace() or token == '\n':
-                yield token, line
-
-    def _read_one_token(self, source_code, index, token):
-        original_index = index
-        if token.startswith("#"):
-            token = "#" + token[1:].strip()
-
-        if token == "#define":
-            while(1):
-                bindex = index + 1
-                index = source_code.find('\n', bindex)
-                if index == -1:
-                    break
-                if not source_code[bindex:index].rstrip().endswith('\\'):
-                    break
-            if index != -1:
-                token = source_code[original_index:index]
-        elif token == '/*':
-            index = source_code.find("*/", index + 2)
-            if index != -1:
-                index += 2
-                token = source_code[original_index:index]
-        elif token in( '//', '#if','#endif'):
-            index = source_code.find('\n', index)
-        elif token == '"' or token == '\'':
-            while(1):
-                index += 1
-                index = source_code.find(token, index)
-                if index == -1:
-                    break
-                if source_code[index - 1] == '\\':
-                    if source_code[index - 2] != '\\':
-                        continue
-                break
-            if index != -1:
-                token = source_code[original_index:index + 1]
-                index = index + 1
-        else:
-            index += len(token)
-        return index, token
+        return [t for t in self.token_pattern.findall(source_code) if not t.isspace() or t == '\n']
 
 generate_tokens = FreeFormattingTokenizer()
 
@@ -814,78 +848,9 @@ class XMLFormatter(object):
         item.appendChild(value4)
         return item
 
-def createCommandLineParser():
-    from optparse import OptionParser
-    parser = OptionParser(version=VERSION)
-    parser.add_option("-v", "--verbose",
-            help="Output in verbose mode (long function name)",
-            action="store_true",
-            dest="verbose",
-            default=False)
-    parser.add_option("-C", "--CCN",
-            help =  "Threshold for cyclomatic complexity number warning. "+
-                    "The default value is %d. Functions with CCN bigger than this number will generate warning" % DEFAULT_CCN_THRESHOLD,
-            action="store",
-            type="int",
-            dest="CCN",
-            default=DEFAULT_CCN_THRESHOLD)
-    parser.add_option("-a", "--arguments",
-            help="Limit for number of parameters",
-            action="store",
-            type="int",
-            dest="arguments",
-            default=100)
-    parser.add_option("-w", "--warnings_only",
-            help="Show warnings only, using clang/gcc's warning format for printing warnings. http://clang.llvm.org/docs/UsersManual.html#cmdoption-fdiagnostics-format",
-            action="store_true",
-            dest="warnings_only",
-            default=False)
-    parser.add_option("-i", "--ignore_warnings",
-            help="If the number of warnings is equal or less than the number, the tool will exit normally, otherwize it will generate error. Useful in makefile when improving legacy code.",
-            action="store",
-            type="int",
-            dest="number",
-            default=0)
-    parser.add_option("-x", "--exclude",
-            help="Exclude files that match this pattern. * matches everything, ? matches any single characoter, \"./folder/*\" exclude everything in the folder, recursively. Multiple patterns can be specified. Don't forget to add \"\" around the pattern.",
-            action="append",
-            dest="exclude",
-            default=[])
-    parser.add_option("-X", "--xml",
-            help="Generate XML in cppncss style instead of the normal tabular output. Useful to generate report in Jenkins server",
-            action="store_true",
-            dest="xml",
-            default=None)
-    parser.add_option("-P", "--no_preprocessor_count",
-            help="By default, a #if will also increase the complexity. Adding this option to ignore them",
-            action="store_true",
-            dest="no_preprocessor_count",
-            default=False)
-    parser.add_option("-t", "--working_threads",
-            help="number of working threads. The default value is 1.",
-            action="store",
-            type="int",
-            dest="working_threads",
-            default=1)
-    parser.add_option("-d", "--find_duplicates",
-            help="find and skip analysis for identical files. Will be made default in the next release",
-            action="store_true",
-            dest="duplicates",
-            default=False)
-    parser.add_option("-e", "--display_fn_end_line",
-            help="display function end line number in addition to start line number. Will be made default in the next release",
-            action="store_true",
-            dest="display_fn_end_line",
-            default=False)
-    parser.add_option("-E", "--extension",
-            help="under construction...", #"Use extension. Can be WordCount.",
-            action="append",
-            dest="extensions",
-            default=[])
+def print_xml(r, options):
+        print (XMLFormatter().xml_output(list(r), options))
 
-    parser.usage = "lizard [options] [PATH or FILE] [PATH] ... "
-    parser.description = __doc__
-    return parser
 
 def mapFilesToAnalyzer(files, fileAnalyzer, working_threads):
     try:
@@ -913,7 +878,7 @@ class FilesFilter(object):
         self.exclude_patterns = exclude_patterns
         self.check_duplicates = check_duplicates
         self.hash_set = set()
-        
+
     def getFileNames(self, paths):
         for SRC_DIR in paths:
             if os.path.isfile(SRC_DIR) :
@@ -954,39 +919,43 @@ class FilesFilter(object):
             self.hash_set.add(fhash)
             return True
 
-def get_extensions(extension_names):
-    return [__import__('lizard' + name).LizardExtension() for name in extension_names]
-
 def analyze(paths, options):
     ''' This is the most important function of lizard.
         It analyze the given paths with the options.
         Can be used directly by other Python application.
     '''
+
+    languages = LanguageChooser(
+            includeHashIfConditions = not options.no_preprocessor_count,
+            switchCasesAsOneCondition = options.switchCasesAsOneCondition,)
+
     files = FilesFilter(options.exclude, options.duplicates).getFileNames(paths)
-    fileAnalyzer = FileAnalyzer(options.extensions, options.no_preprocessor_count)
-    r = mapFilesToAnalyzer(files, fileAnalyzer, options.working_threads)
-    return r
+    fileAnalyzer = FileAnalyzer(languages, options.extensions)
+    return mapFilesToAnalyzer(files, fileAnalyzer, options.working_threads)
 
-def get_whitelist():
-    try:
-        with open('whitelizard.txt', mode='r') as whitelizard:
-            return whitelizard.read()
-    except:
-        return ''
+def parse_args(argv):
+
+    def get_extensions(extension_names):
+        return [__import__('lizard' + name).LizardExtension() for name in extension_names]
+
+    def get_whitelist():
+        try:
+            with open('whitelizard.txt', mode='r') as whitelizard:
+                return whitelizard.read()
+        except:
+            return ''
 
-def lizard_main(argv):
     options, args = createCommandLineParser().parse_args(args=argv)
     options.whitelist = get_whitelist()
     options.extensions = get_extensions(options.extensions)
     paths = ["."] if len(args) == 1 else args[1:]
-    r = analyze(paths, options)
-    if options.xml:
-        print (XMLFormatter().xml_output(list(r), options))
-    else:
-        print_result(r, options)
+    return paths, options
 
-def main():
-    lizard_main(sys.argv)
+def lizard_main(argv =  sys.argv):
+    paths, options = parse_args(argv)
+    printer = print_xml if options.xml else print_result
+    r = analyze(paths, options)
+    printer(r, options)
 
 if __name__ == "__main__":
-    main()
+    lizard_main()
